# -*- coding: utf-8 -*-

"""
Module latticeadaptor.mappings 
=================================================================

A module containing mappings between elements and attributes for the 
different lattice formats used:
* MadX
* Elegant
* Tracy

"""



# -*- coding: utf-8 -*-

"""
Module latticeadaptor.utils 
=================================================================

A module containg some helper functions.

"""

from typing import List, Tuple, Union

import pandas as pd


def is_number(s: str) -> bool:
    """Method to check if a value is a number or not.

    Parameters
    ----------
    s : str
        string to test

    Returns
    -------
    bool
        true or false
    """
    try:
        float(s)
        return True
    except ValueError:
        return False


def filter_family(df: pd.DataFrame, fam: str) -> pd.DataFrame:
    """Filter the dataframe by family name.

    Parameters
    ----------
    df : pd.DataFrame
        dataframe to filter
    fam : str
        family to filter on

    Returns
    -------
    pd.DataFrame
        copy of filtered dataframe
    """
    _filter = df.family == fam

    return df.loc[_filter].copy().reset_index(drop=True)


def rotate(l: list, n: int) -> list:
    """Method to rotate elements in a list.

    Parameters
    ----------
    l : list
        input list
    n : int
        number of element to rotate over

    Returns
    -------
    list
        rotated list
    """
    return l[-n:] + l[:-n]


def delete_first_line(file: str):
    """Method to delete the first line in a file.

    Parameters
    ----------
    file : str
        filename
    """
    with open(file, "r") as fin:
        data = fin.read().splitlines(True)
    with open(file, "w") as fout:
        fout.writelines(data[1:])


def highlight_cells(
    data: pd.DataFrame, _list: list = [], color: str = "yellow"
) -> Union[pd.DataFrame, list]:
    """Method that can be used by pd.style.apply for styling single cells.

    Parameters
    ----------
    data : pd.DataFrame
        input dataframe
    _list : list, optional
        selection value list, by default []
    color : str, optional
        color to use to highlight, by default "yellow"

    Returns
    -------
    Union[pd.DataFrame, list]
        list or dataframe to be used by the style.apply method of pandas
    """
    attr = "background-color: {}".format(color)

    # series of dataframe
    if data.ndim == 1:
        is_sel = data.isin(_list)
        return [attr if v else "" for v in is_sel]
    else:
        is_sel = data.isin(_list)
        return pd.DataFrame(np.where(is_sel, attr, ""), index=data.index, columns=data.columns)


def highlight_row(data: pd.DataFrame, _list: list, column: str, color: str = "yellow") -> list:
    """Method to highlight rows in a dataframe with a given color.

    Parameters
    ----------
    data : pd.DataFrame
        input dataframe
    _list : list
        selection list
    column : str
        column to apply selection to
    color : str, optional
        color to use, by default "yellow"

    Returns
    -------
    list
        list to be used by the style.apply method of pandas
    """
    attr = "background-color: {}".format(color)
    is_sel = pd.Series(data=False, index=data.index)
    is_sel[column] = data.loc[column].isin(_list)

    return [attr if is_sel.any() else "" for v in is_sel]


def display_more(df: pd.DataFrame, maxrows: int = 300, maxcols: int = 100):
    """
    Show more rows and columns of a dataframe.
    """
    with pd.option_context("display.max_rows", maxrows, "display.max_columns", maxcols):
        display(df)


# ==============================================================================
# Lattice tools
# ==============================================================================
def dipole_split_angles_to_dict(
    dipole_name: str,
    dipole_len: float,
    dipole_bend_angle_rad: float,
    angle_list: list,
    verbose: bool = True,
) -> dict:
    """Method to generate a dictionary that is used by the `split_dipoles` method to split
    the dipoles over the angle list given in the input. Verbose allows to print the output
    for debugging.

    Parameters
    ----------
    dipole_name : str
        dipole name of the dipole to be split
    dipole_len : float
        dipole length
    dipole_bend_angle_rad : float
        dipole bending angle
    angle_list : list
        list of angles to split over
    verbose : bool, optional
        print output, by default True

    Returns
    -------
    dict
        Dictionary containing the split angles in rad and split lengths. Main key is the dipole name, subkeys are
        'lengths' and 'angles'

    IMPORTANT NOTE:
    ---------------
    Auto adds half angle split and final angle for full magnet.
    """
    _dict = {dipole_name: dict()}

    dipole_bend_angle_deg = np.rad2deg(dipole_bend_angle_rad)
    dipole_bend_radius = dipole_len / dipole_bend_angle_rad

    if verbose:
        print("Dipole length [m]       : {:12.6f}".format(dipole_len))
        print("Dipole Bend Angle [rad] : {:12.6f}".format(dipole_bend_angle_rad))
        print("Dipole Bend Angle [deg] : {:12.6f}".format(dipole_bend_angle_deg))
        print("Dipole Bend Radius [m]  : {:12.6f}".format(dipole_bend_radius))
        print()
        print()

    split_angles_BM_deg = np.array(
        sorted(angle_list + [dipole_bend_angle_deg / 2, dipole_bend_angle_deg])
    )
    split_angles_BM_cum_rad = np.deg2rad(split_angles_BM_deg)
    split_angles_BM_rad = np.r_[split_angles_BM_cum_rad[0], np.diff(split_angles_BM_cum_rad)]
    split_lengths_cum_BM = split_angles_BM_cum_rad * dipole_bend_radius
    split_lengths_BM = np.r_[split_lengths_cum_BM[0], np.diff(split_lengths_cum_BM)]

    if verbose:
        print(
            "BM splitting angles               [deg] : {}".format(
                "".join(["{:12.6f}".format(a) for a in split_angles_BM_deg])
            )
        )
        print(
            "BM splitting angles  - cumulative [rad] : {}".format(
                "".join(["{:12.6f}".format(a) for a in split_angles_BM_cum_rad])
            )
        )
        print(
            "BM splitting angles  - individual [rad] : {}".format(
                "".join(["{:12.6f}".format(a) for a in split_angles_BM_rad])
            )
        )
        print(
            "BM splitting lengths - cumulative [m]   : {}".format(
                "".join(["{:12.6f}".format(a) for a in split_lengths_cum_BM])
            )
        )
        print(
            "BM splitting lengths - individual [m]   : {}".format(
                "".join(["{:12.6f}".format(a) for a in split_lengths_BM])
            )
        )
        print()

    _dict[dipole_name] = {"lengths": split_lengths_BM, "angles": split_angles_BM_rad}

    return _dict


def split_dipoles(df: pd.DataFrame, _dict: dict, halfbendangle: float) -> pd.DataFrame:
    """Method to split the dipole given in the input.

    Parameters
    ----------
    df : pd.DataFrame
        input table of dipoles
    _dict : dict
        splitting dictionary
    halfbendangle : float
        half bend angle of the dipole to split

    Returns
    -------
    pd.DataFrame
        update table with the dipole split.
    """
    # init output
    newdf = pd.DataFrame()

    # loop over the dipoles
    for j, row in df.iterrows():
        lengths = _dict[row["name"]]["lengths"]
        angles = _dict[row["name"]]["angles"]

        # calculate the center positions of the splits
        end_pos = lengths.cumsum() + row["at"] - row.L / 2
        center_pos = end_pos - (lengths / 2)

        # count splits per magnet
        aports = 1
        bports = 1
        cum_angle = 0

        # loop over the splitting angles for this specific dipole
        for i, (angle, l, pos) in enumerate(zip(angles, lengths, center_pos)):
            # for each angle create a new seq table entry
            newrow = row.copy()
            cum_angle += angle

            # naming depends on magnet number in the sector
            if "BM1" in row["name"]:
                M = 1
            else:
                M = 2

            # add marker
            markerrow = row.copy()
            markerrow.family = "MARKER"
            markerrow.L = 0.000000
            markerrow = markerrow.drop(labels=["E1", "E2", "K1", "K2", "ANGLE"], errors="ignore")
            # naming
            # beam ports A in first half of the magnet
            # beam ports B in second half of the magnet
            # number each per split number
            if cum_angle - halfbendangle < -1e-6:
                name = row["name"] + "1_{}_deg".format(
                    ("{:2.2f}".format(np.rad2deg(cum_angle))).replace(".", "p").strip()
                )
                markerrow["name"] = "MBEAMPORT_{}A{}".format(M, aports)
                aports += 1
            elif abs(cum_angle - halfbendangle) < 1e-6:
                name = row["name"] + "1_{}_deg".format(
                    ("{:2.2f}".format(np.rad2deg(cum_angle))).replace(".", "p").strip()
                )
                markerrow["name"] = (
                    "M"
                    + row["name"]
                    + "_MIDDLE".format(
                        ("{:2.2f}".format(np.rad2deg(cum_angle))).replace(".", "p").strip()
                    )
                )
            else:
                name = row["name"] + "2_{}_deg".format(
                    ("{:2.2f}".format(np.rad2deg(cum_angle))).replace(".", "p").strip()
                )
                markerrow["name"] = "MBEAMPORT_{}B{}".format(M, bports)
                bports += 1

            newrow["name"] = name

            # updating angles
            newrow.ANGLE = angle

            # updating lengths
            newrow.L = l

            # updating center position
            newrow.pos = pos
            markerrow["pos"] = pos + l / 2

            # if at is already in columns update it
            if "at" in newrow.index:
                newrow["at"] = pos

            # if at is already in columns update it
            if "at" in markerrow.index:
                markerrow["at"] = markerrow["pos"]

            # update E1 E2
            if i != 0:
                newrow.E1 = 0.000000

                if i != len(angles) - 1:
                    newrow.E2 = 0.000000
            else:
                newrow.E2 = 0.000000

            # add marker only if not at end of magnet
            if abs(cum_angle - 2 * halfbendangle) > 1e-6:
                newdf = newdf.append(markerrow)

            newdf = newdf.append(newrow)

    return newdf.reset_index(drop=True)


def compare_settings_dicts(dc1: dict, dc2: dict, threshold: float = 1.0) -> None:
    """Method to compare lattice setting dicts.

    Parameters
    ----------
    dc1 : dict
        settings in first lattice
    dc2 : dict
        settings in second lattice
    threshold : float, optional
        threshold to use for issuing warnings, by default 1.0
    """
    combinedc = defaultdict(list)
    for k, v in chain(dc1.items(), dc2.items()):
        combinedc[k].append(v)

    for k, v in combinedc.items():
        if len(v) > 1:
            if v[0] != v[1]:
                if abs(v[1] - v[0]) > threshold:
                    print(colored("WARNING STRONG CHANGE", "red"))
                print(
                    colored(
                        "{:10} {:16.12f} {:16.12f} {:16.12f}".format(k, v[0], v[1], v[1] - v[0]),
                        "yellow",
                    )
                )
            else:
                print(
                    colored(
                        "{:10} {:16.12f} {:16.12f}".format(k, v[0], v[1], v[1] - v[0]), "green"
                    )
                )


def print_twiss_summ(tw):
    """Method to pretty print madx twiss summary generated by cpymadx

    Parameters:
    -----------
    tw : cpymad twiss object
        twiss
    """
    from pprint import pprint

    try:
        dc = [
            "{:15}:{:15.6e}".format(k, v)
            if isinstance(v, float)
            else "{:15}:{:>15}".format(k, v.strip())
            for k, v in dict(tw.summary).items()
        ]
    except:
        dc = [
            "{:15}:{:15.6e}".format(k, v)
            if isinstance(v, float)
            else "{:15}:{:>15}".format(k, v.strip())
            for k, v in dict(tw).items()
        ]
    pprint(dc)


# ==============================================================================
# MADX tools
# ==============================================================================
def install_start_end_marker(name: str, length: float) -> str:
    """Method to add start and end marker to the lattice.

    Parameters
    ----------
    name : str
        lattice name
    length : float
        length of the lattice

    Returns
    -------
    str
        MADX install string.
    """
    # define  start and end marker
    text = "{:12}: {:12};\n".format("MSTART", "MARKER")
    text += "{:12}: {:12};\n\n".format("MEND", "MARKER")

    # start sequence edit
    text += "USE, SEQUENCE={};\n".format(name)
    text += "SEQEDIT, SEQUENCE = {};  \nFLATTEN;\n".format(name)

    # install start and end marker
    line = "INSTALL, ELEMENT = {:16}, AT = {:12.6f};\n".format("MSTART", 0.00000)
    text += line
    line = "INSTALL, ELEMENT = {:16}, AT = {:12.6f};\n".format("MEND", length)
    text += line

    # end sequence edit
    text += "FLATTEN;\nENDEDIT;"

    return text


# ==============================================================================
# Plot tools
# ==============================================================================
from collections import defaultdict
from itertools import chain

import matplotlib
import matplotlib.patches as mpatches
import matplotlib.path as mpath
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from cycler import cycler
from matplotlib import pyplot as plot
from matplotlib.collections import PatchCollection
from matplotlib.figure import Figure
from mpl_toolkits.mplot3d import Axes3D
from termcolor import colored

from ..parsers import parse_from_madx_sequence_string


def draw_brace(ax, xspan, text, yshift=0.0):
    """Draws an annotated brace on the x axes."""
    xmin, xmax = xspan
    xspan = xmax - xmin
    ax_xmin, ax_xmax = ax.get_xlim()
    xax_span = ax_xmax - ax_xmin
    ymin, ymax = ax.get_ylim()
    yspan = ymax - ymin
    resolution = int(xspan / xax_span * 100) * 2 + 1  # guaranteed uneven
    beta = 300.0 / xax_span  # the higher this is, the smaller the radius

    x = np.linspace(xmin, xmax, resolution)
    x_half = x[: resolution // 2 + 1]
    y_half_brace = 1 / (1.0 + np.exp(-beta * (x_half - x_half[0]))) + 1 / (
        1.0 + np.exp(-beta * (x_half - x_half[-1]))
    )
    y = np.concatenate((y_half_brace, y_half_brace[-2::-1]))
    y = ymin + +yshift + (0.05 * y - 0.01) * yspan  # adjust vertical position

    ax.autoscale(False)
    ax.plot(x, y, color="black", lw=1)

    ax.text((xmax + xmin) / 2.0, ymin + 0.07 * yspan, text, ha="center", va="bottom")


def Beamlinegraph_compare_from_seq_files(
    seqfile1: str, seqfile2: str, start: float = 0.0, stop: Union[float, None] = None
):
    """Method to compare the location of beam line elements where the positions
    are extracted from MADX sequence files.

    Parameters
    ----------
    seqfile1 : str
        sequence file 1
    seqfile2 : str
        sequence file 2
    start : float, optional
        start postion for plotting, by default 0.0
    stop : [type], optional
        end position for plotting, by default Union[float, None]

    Returns:
    --------
    Tuple
        pyplot obj and ax obj
    """
    _REQUIRED_COLUMNS = ["pos", "name", "L"]
    _RECTANGLE_ELEMENTS = [
        "SBEND",
        "RBEND",
        "KICKER",
        "VKICKER",
        "HKICKER",
        "QUADRUPOLE",
        "SEXTUPOLE",
        "RFCAVITY",
        "DRIFT",
    ]
    _COLMAP = {
        "SBEND": "#0099FF",
        "RBEND": "#0099FF",
        "KICKER": "#0099FF",
        "VKICKER": "#0099FF",
        "HKICKER": "#0099FF",
        "QUADRUPOLE": "green",
        "SEXTUPOLE": "#FF99FF",
        "RFCAVITY": "#08bad1",
        "MARKER": "red",
        "MONITOR": "black",
        "DIPEDGE": "blue",
        "DRIFT": "#D0D0D0",
    }

    with open(seqfile1, "r") as f:
        seqfilestr1 = f.read()

    with open(seqfile2, "r") as f:
        seqfilestr2 = f.read()

    name1, len1, table1 = parse_from_madx_sequence_string(seqfilestr1)
    name2, len2, table2 = parse_from_madx_sequence_string(seqfilestr2)

    # check if columns are ok
    for c in _REQUIRED_COLUMNS:
        assert c in table1.columns
        assert c in table2.columns

    # find range to plot
    if stop is None:
        stop = max(len1, len2)
        # idxmax_pos = table1["pos"].idxmax()
        # stop = table1.loc[idxmax_pos, "pos"] + table1.loc[idxmax_pos, "L"]
    else:
        table1 = table1.loc[table1["pos"].between(start, stop)]
        table2 = table2.loc[table2["pos"].between(start, stop)]

    _ = plt.figure(figsize=(16, 6))
    axis = plt.gca()

    offset_array = np.array([0.0, -0.5])

    for _, row in table1.iterrows():
        col = _COLMAP[row.family.upper()]

        if row.family in _RECTANGLE_ELEMENTS:
            axis.add_patch(
                mpatches.Rectangle(
                    np.array([row.pos - row.L / 2, 0.0]) + offset_array,
                    row.L,
                    0.33,
                    color=col,
                    alpha=1.0,
                )
            )
            axis.vlines(
                row.pos - row.L / 2,
                offset_array[1],
                0,
                linestyle="dashed",
                color="gray",
                linewidth=1,
            )
            axis.vlines(
                row.pos + row.L / 2,
                offset_array[1],
                0,
                linestyle="dashed",
                color="gray",
                linewidth=1,
            )

        else:
            Path = mpath.Path
            h = 1.0
            offs_mon = np.array([0.0, 0.03])

            verts = np.array(
                [
                    (0, h),
                    (0, -h),
                ]
            )

            codes = [Path.MOVETO, Path.LINETO]

            verts += offs_mon

            path = mpath.Path(verts + offset_array + np.array([row.pos, 0.0]), codes)
            patch = mpatches.PathPatch(path, color=col, lw=1, alpha=1.0 * 0.5)
            axis.add_patch(patch)

        axis.annotate(
            row.family + ": " + row["name"],
            xy=(row.pos, 0),
            # xycoords='data',
            xytext=(row.pos, 0) + offset_array,
            # textcoords='data',
            horizontalalignment="left",
            # arrowprops=dict(arrowstyle="simple",),# connectionstyle="arc3,rad=+0.2"),
            # bbox=dict(boxstyle="round", facecolor="w", edgecolor="0.5", alpha=0.9),
            fontsize=8,
            rotation=90,
        )

    offset_array = np.array([0.0, 0.5])

    for _, row in table2.iterrows():
        col = _COLMAP[row.family.upper()]

        if row.family in _RECTANGLE_ELEMENTS:
            axis.add_patch(
                mpatches.Rectangle(
                    np.array([row.pos - row.L / 2, 0.0]) + offset_array,
                    row.L,
                    0.33,
                    color=col,
                    alpha=1.0,
                )
            )
            axis.vlines(
                row.pos - row.L / 2,
                offset_array[1],
                0,
                linestyle="dashed",
                color="red",
                linewidth=1,
            )
            axis.vlines(
                row.pos + row.L / 2,
                offset_array[1],
                0,
                linestyle="dashed",
                color="red",
                linewidth=1,
            )
        else:
            Path = mpath.Path
            h = 1.0
            offs_mon = np.array([0.0, 0.03])

            verts = np.array(
                [
                    (0, h),
                    (0, -h),
                ]
            )

            codes = [Path.MOVETO, Path.LINETO]

            verts += offs_mon

            path = mpath.Path(verts + offset_array + np.array([row.pos, 0.0]), codes)
            patch = mpatches.PathPatch(path, color=col, lw=1, alpha=1.0 * 0.5)
            axis.add_patch(patch)

        axis.annotate(
            row.family + ": " + row["name"],
            xy=(row.pos, 0),
            # xycoords='data',
            xytext=(row.pos, 0) + offset_array,
            # textcoords='data',
            horizontalalignment="left",
            # arrowprops=dict(arrowstyle="simple",),# connectionstyle="arc3,rad=+0.2"),
            # bbox=dict(boxstyle="round", facecolor="w", edgecolor="0.5", alpha=0.9),
            fontsize=8,
            rotation=90,
        )

    plt.xlim(start, stop)
    plt.ylim(-1.1, 1.1)
    plt.xlabel("S[m]")
    #     plt.grid()
    return plt, axis


def Beamlinegraph_from_seq_file(
    seqfile: str,
    start: float = 0.0,
    stop: Union[float, None] = None,
    offset_array: Union[list, np.array] = [0.0, 0.0],
    anno: bool = True,
    size: Tuple[int] = (12, 6),
):
    """Method to generate beam line graph plot from MADX seq file.

    Parameters
    ----------
    seqfile : str
        MADX sequence file
    start : float, optional
        start position of the plot, by default 0.0
    stop : Union[float, None], optional
        end position of the plot, by default None
    offset_array : Union[list, np.array], optional
        beamline graph offset, by default [0.0, 0.0]
    anno : bool, optional
        annotate elements, by default True
    size : Tuple[int], optional
        plotsize, by default (12, 6)

    Returns
    -------
    Tuple
        pyplot obj and ax obj
    """
    _REQUIRED_COLUMNS = ["pos", "name", "L"]
    _RECTANGLE_ELEMENTS = ["SBEND", "RBEND", "KICKER", "VKICKER", "HKICKER", "DRIFT"]
    _MIN_HEIGTH = 0.1
    _MAX_HEIGTH = 10.0

    with open(seqfile, "r") as f:
        seqfilestr = f.read()

    name, length, table = parse_from_madx_sequence_string(seqfilestr)

    # check if columns are ok
    for c in _REQUIRED_COLUMNS:
        assert c in table.columns

    # find range to plot
    if stop is None:
        # idxmax_pos = table["pos"].idxmax()
        # stop = table.loc[idxmax_pos, "pos"] + table.loc[idxmax_pos, "L"]
        stop = length
    else:
        table = table.loc[table["pos"].between(start, stop)]

    #    print(start, stop)

    element_families = table.family.unique()

    # determine max plot heights
    if any(i in element_families for i in _RECTANGLE_ELEMENTS):
        angle_max = 1.0
        if ("ANGLE" in table.columns) and not all(table.ANGLE.isna()):
            angle_max = abs(table.ANGLE.max())
        else:
            table["ANGLE"] = 0.0

    if "QUADRUPOLE" in element_families:
        k_max = abs(table.K1.max())

    _ = plt.figure(figsize=size)
    axis = plt.gca()

    for _, row in table.iterrows():
        if row.family in _RECTANGLE_ELEMENTS:
            axis.add_patch(
                mpatches.Rectangle(
                    np.array([row.pos - row.L / 2, 0.0]) + offset_array,
                    row.L,
                    np.sign(row.ANGLE) * _MIN_HEIGTH + row.ANGLE / angle_max * (1 - _MIN_HEIGTH),
                    color="#0099FF",
                    alpha=1.0,
                )
            )

        elif row.family.upper() == "QUADRUPOLE":
            if row.K1 >= 0:
                # FOCUSSING QUAD
                axis.add_patch(
                    mpatches.Ellipse(
                        offset_array + np.array([row.pos, 0.0]),
                        row.L,
                        _MIN_HEIGTH + abs(row.K1 / k_max) * (1 - _MIN_HEIGTH),
                        color="green",
                        alpha=1.0,
                    )
                )
            else:
                Path = mpath.Path
                h = _MIN_HEIGTH + abs(row.K1 / k_max) * (
                    1 - _MIN_HEIGTH
                )  # abs(row.K1 / k_max) + _MIN_HEIGTH / 2
                dx = row.L
                verts = np.array(
                    [
                        (dx, h),
                        (-dx, h),
                        (-dx / 4, 0),
                        (-dx, -h),
                        (dx, -h),
                        (dx / 4, 0),
                        (dx, h),
                    ]
                )

                codes = [
                    Path.MOVETO,
                    Path.LINETO,
                    Path.CURVE3,
                    Path.LINETO,
                    Path.LINETO,
                    Path.CURVE3,
                    Path.CURVE3,
                ]

                path = mpath.Path(verts + offset_array + np.array([row.pos, 0.0]), codes)
                patch = mpatches.PathPatch(path, color="green", alpha=1.0)
                axis.add_patch(patch)
        elif row.family in ["SEXTUPOLE"]:
            axis.add_patch(
                mpatches.RegularPolygon(
                    np.array([row.pos, 0.0]) + offset_array,
                    6,
                    row.L / 2,
                    color="#FF99FF",
                    alpha=1.0,
                )
            )
        elif row.family in ["OCTUPOLE"]:
            axis.add_patch(
                mpatches.RegularPolygon(
                    np.array([row.pos, 0.0]) + offset_array,
                    8,
                    row.L / 2,
                    color="#FF99FF",
                    alpha=1.0,
                )
            )
        elif row.family in ["MONITOR"]:
            Path = mpath.Path
            h = 0.25
            offs_mon = np.array([0.0, 0.03])

            verts = np.array([(0, h), (0, -h), (h, 0), (-h, 0)])

            codes = [Path.MOVETO, Path.LINETO, Path.MOVETO, Path.LINETO]

            verts += offs_mon

            path = mpath.Path(verts + offset_array + np.array([row.pos, 0.0]), codes)
            patch = mpatches.PathPatch(path, color="black", lw=2, alpha=1.0 * 0.5)
            axis.add_patch(patch)
            axis.add_patch(
                mpatches.Circle(
                    offset_array + offs_mon + np.array([row.pos, 0.0]),
                    h / 2,
                    color="black",
                    alpha=1.0 * 0.25,
                )
            )
        elif row.family in ["MARKER"]:
            Path = mpath.Path
            h = 1.0
            offs_mon = np.array([0.0, 0.03])

            verts = np.array(
                [
                    (0, h),
                    (0, -h),
                ]
            )

            codes = [Path.MOVETO, Path.LINETO]

            verts += offs_mon

            path = mpath.Path(verts + offset_array + np.array([row.pos, 0.0]), codes)
            patch = mpatches.PathPatch(path, color="red", lw=2, alpha=1.0 * 0.5)
            axis.add_patch(patch)

        elif row.family in ["CAVITY"]:
            axis.add_patch(
                mpatches.Ellipse(
                    np.array([row.pos - row.L, 0.0]),
                    row.L,
                    2 * row.L,
                    angle=0,
                    color="#08bad1",
                    alpha=1 * 0.5,
                )
            )

        elif row.family in ["RFMODE"]:
            axis.add_patch(
                mpatches.Ellipse(
                    np.array([row.pos - row.L / 2, 0.0]),
                    row.L,
                    10 * row.L,
                    angle=0,
                    color="#f2973d",
                    alpha=1 * 0.5,
                )
            )

        elif row.family in ["DRIFT"]:
            axis.add_patch(
                mpatches.Rectangle(
                    np.array([row.pos - row.L / 2, -_MIN_HEIGTH]),
                    row.L,
                    2 * _MIN_HEIGTH,
                    color="black",
                    alpha=1 * 0.4,
                )
            )

        if anno:
            annotation = axis.annotate(
                row.family + ": " + row["name"],
                xy=(row.pos + offset_array[0], 0 + offset_array[1]),
                # xycoords='data',
                xytext=(row.pos + offset_array[0], 0 + offset_array[1]),
                # textcoords='data',
                horizontalalignment="left",
                # arrowprops=dict(arrowstyle="simple",),# connectionstyle="arc3,rad=+0.2"),
                # bbox=dict(boxstyle="round", facecolor="w", edgecolor="0.5", alpha=0.9),
                fontsize=8,
                rotation=90,
            )
    plt.xlim(start, stop)
    plt.ylim(-1.1, 1.1)
    plt.xlabel("S[m]")
    #     plt.grid()
    return plt, axis


def twissplot(
    tw, cols=["betx", "bety", "dx"], cpymadtwiss=True, beamlinegraph=False, *args, **kwargs
):
    """
    Method to plot columns from the twiss table output using cpymadtwiss.
    """
    if cpymadtwiss:
        if beamlinegraph:
            plot, ax = Beamlinegraph_from_seq_file(
                kwargs.get("sequence"),
                offset_array=kwargs.get("offset_array", np.array([0.0, 0.0])),
                start=kwargs.get("start", 0.0),
                stop=kwargs.get("stop", None),
                anno=kwargs.get("anno", False),
            )
        else:
            _ = plt.figure(figsize=kwargs.get("size", (12, 6)))
            plot = plt.gcf()
            ax = plt.gca()

        linestyle_cycler = ["-", "--", ":", "-."]
        if isinstance(tw, list):
            for i, twi in enumerate(tw):
                ax.set_prop_cycle(
                    cycler(
                        "color",
                        [
                            "#1f77b4",
                            "#ff7f0e",
                            "#2ca02c",
                            "#d62728",
                            "#9467bd",
                            "#8c564b",
                            "#e377c2",
                            "#7f7f7f",
                            "#bcbd22",
                            "#17becf",
                        ],
                    )
                    * cycler("linestyle", [linestyle_cycler[i]])
                )
                # ax.rc('axes', prop_cycle=linestyle_cycler)
                for col in cols:
                    ax.plot(twi.get("s"), twi.get(col), label=col + "_{}".format(i))
        else:
            for col in cols:
                ax.plot(tw.get("s"), tw.get(col), label=col)

        ax.legend()
        ax.relim()
        ax.autoscale()
        return plot, ax

    else:
        print("Not implemented yet!!!")
        pass



# -*- coding: utf-8 -*-

"""
Module latticeadaptor.core 
=================================================================

A module containing the main class to operate on accelerator lattices.

"""

import os

# your imports here ...
import queue
import re
from copy import deepcopy
from typing import Tuple

import numpy as np
import pandas as pd
from latticeconstructor.core import LatticeBuilderLine
from latticeconstructor.parse import parse_from_string

from .parsers import (
    _parse_table_to_madx_definitions,
    parse_from_madx_sequence_file,
    parse_from_madx_sequence_string,
    parse_table_to_elegant_file,
    parse_table_to_elegant_string,
    parse_table_to_madx_install_str,
    parse_table_to_madx_remove_str,
    parse_table_to_madx_sequence_file,
    parse_table_to_madx_sequence_string,
    parse_table_to_tracy_file,
    parse_table_to_tracy_string,
    save_string,
)
from .utils import install_start_end_marker


class LatticeAdaptor:
    """Class to convert lattices."""

    def __init__(self, **kwargs):
        # roll back
        self.history = queue.LifoQueue()
        self.name = kwargs.get("name", None)
        self.len = kwargs.get("len", 0.0)
        self._table = None
        self.table = kwargs.get("table", None)
        self.filename = kwargs.get("file", None)
        self.inputstr = kwargs.get("string", None)
        self.builder = LatticeBuilderLine()

    @property
    def table(self):
        return self._table

    @table.setter
    def table(self, value):
        # roll back
        self.history.put((deepcopy(self.name), deepcopy(self.len), deepcopy(self.table)))

        self._table = value

    def load_from_file(self, filename: str, ftype: str = "lte") -> None:
        """Load data from file.

        Parameters
        ----------
        filename : str
            filename of the file to load lattice from
        ftype : str, optional
            lattice format to load from, currenlty lte and madx allowed, by default "lte"
        """

        # use the latticebuilder to build the table
        self.builder.load_from_file(filename, ftype)
        self.builder.build_table()

        # extract the relevant info
        # Better way?
        try:
            self.name = self.builder.name
        except Exception:
            self.name = "ring"
            print("Lattice name has been autoset - check if value is ok.")
        self.table = self.builder.table

        # length is last element center pos + half the  length
        with open(filename, "r") as f:
            txt = f.read()

        length = re.findall("sequence\s*,\s*\s*[lL]\s*[:]?=\s*(\d+[.]?\d+)", txt)

        if (ftype == "madx") and (len(length) != 0):
            self.len = length[0]
        if not self.len:
            self.len = (
                self.table.tail(1)["at"].values[-1] + self.table.tail(1)["L"].values[-1] / 2.0
            )
            print(
                "Length has been autoset to {} - check if value is ok - otherwise update it.".format(
                    self.len
                )
            )

    def load_from_string(self, string: str, ftype: str = "lte") -> None:
        """Load data from string.

        Parameters
        ----------
        string : str
            string input
        ftype : str, optional
            format of string, currenlty supported are lte and madx, by default "lte"
        """
        # roll back
        self.history.put((deepcopy(self.name), deepcopy(self.len), deepcopy(self.table)))

        # write to temp file
        with open("templatticestringfile.tmp", "w") as f:
            f.write(string)

        # use load from file method
        self.load_from_file("templatticestringfile.tmp", ftype)

    def parse_table_to_madx_sequence_string(self) -> str:
        """Parse table to MADX sequence and return it as a string.

        Returns
        -------
        str
            MADX sequence
        """
        return parse_table_to_madx_sequence_string(self.name, self.len, self.table)

    def parse_table_to_madx_sequence_file(self, filename: str) -> None:
        """Parse table to MADX sequence and write it to file.

        Parameters
        ----------
        filename : str
            filename where the seq will be written to.
        """
        parse_table_to_madx_sequence_file(self.name, self.len, self.table, filename)

    def parse_table_to_elegant_string(self) -> str:
        """Parse table to Elegant lattice and return it as a string.

        Returns
        -------
        str
            Elegant Lattice
        """
        self.add_drifts()
        return parse_table_to_elegant_string(self.name, self.table)

    def parse_table_to_elegant_file(self, filename: str) -> None:
        """Parse table to Elegant lattice and write it to file.

        Parameters
        ----------
        filename : str
            filename where the seq will be written to.
        """
        self.add_drifts()

        parse_table_to_elegant_file(self.name, self.table, filename)

    def parse_table_to_tracy_string(self) -> str:
        """Parse table to Tracy lattice and return it as a string.

        Returns
        -------
        str
            Tracy Lattice
        """
        return parse_table_to_tracy_string(self.name, self.table)

    def parse_table_to_tracy_file(self, filename: str) -> None:
        """Parse table to Tracy lattice and write it to file.

        Parameters
        ----------
        filename : str
            filename where the seq will be written to.
        """
        parse_table_to_tracy_file(self.name, self.table, filename)

    def madx_sequence_add_start_end_marker_string(self) -> str:
        """Return MADX string to install marker at start and end of the lattice.

        Returns
        -------
        str
            MADX install string that can be run with cpymad.
        """
        return install_start_end_marker(self.name, self.len)

    def parse_table_to_madx_install_str(self) -> str:
        """Method to generate a MADX install element string based on the table.
        This string can be used by cpymad to install new elements.

        Returns
        -------
        str
            Install element string input for MADX.
        """
        return parse_table_to_madx_install_str(self.name, self.table)

    def parse_table_to_madx_remove_str(self) -> str:
        """Method to generate a MADX remove element string based on the table.
        This string can be used by cpymad to remove elements.add()

        Returns
        -------
        str
            Remove element string input for MADX.
        """
        return parse_table_to_madx_remove_str(self.name, self.table)

    def madx_sequence_save_string(self, filename: str) -> str:
        """Method to generate string input for MADX to save the lattice
        to sequence.add()

        Parameters
        ----------
        filename : str
            filename of where to write the sequence to.

        Returns
        -------
        str
            save sequence string.
        """
        return "SAVE, SEQUENCE={}, file='{}';".format(self.name, filename)

    def add_drifts(self):
        """Method to add back drifts to sequence."""
        self.history.put((deepcopy(self.name), deepcopy(self.len), deepcopy(self.table)))

        df = self.table.copy()
        df.reset_index(inplace=True, drop=True)
        name = "D"
        family = "DRIFT"

        df.loc[df.L.isna(), "L"] = 0
        if "pos" not in df.columns:
            df["pos"] = df["at"]
        newrows = []
        ndrift = 0
        for i, row in df.iterrows():
            # add the row
            newrows.append(pd.DataFrame(row).T)

            # check if next row
            if i < len(df) - 1:
                # check if next row pos is not equal to the current
                nextrow = df.loc[i + 1]
                # print(
                #    row["pos"],
                #    nextrow["pos"],
                #    nextrow["pos"] > row.pos,
                #    nextrow["pos"] - (nextrow["L"] / 2.0) > row.pos + row.L / 2.0,
                # )
                if nextrow["pos"] - (nextrow["L"] / 2.0) > row.pos + row.L / 2.0:
                    ndrift += 1
                    newrow = {}
                    newrow["name"] = name + str(ndrift)
                    newrow["family"] = family
                    newrow["L"] = np.round(
                        (nextrow["pos"] - nextrow["L"] / 2.0) - (row["pos"] + row["L"] / 2.0), 6
                    )
                    newrow["pos"] = (row["pos"] + row["L"] / 2.0) + (newrow["L"] / 2.0)
                    newrows.append(pd.Series(newrow).to_frame().T)

        # if lattice length is longer than end of last element there is still a drift
        if nextrow["pos"] + nextrow["L"] / 2.0 < self.len:
            newrow = {}
            newrow["name"] = name + str(ndrift)
            newrow["family"] = family
            newrow["L"] = np.round(self.len - nextrow["pos"], 6)
            newrow["pos"] = (row["pos"] + row["L"] / 2.0) + (newrow["L"] / 2.0)
            newrows.append(pd.Series(newrow).to_frame().T)

        self.table = (pd.concat(newrows)).reset_index(drop=True)

    def parse_table_to_madx_line_string(self) -> str:
        """Method to convert the table to a MADX line definition lattice.

        Returns
        -------
        str
            MADX lattice definition string.
        """
        self.add_drifts()
        defstr = _parse_table_to_madx_definitions(self.table)
        linestr = "{}: LINE=({});".format(
            self.name,
            ",\n\t\t".join(
                [",".join(c) for c in list(self.chunks(self.table.name.to_list(), 20))]
            ),
        )
        return defstr + "\n\n" + linestr

    @staticmethod
    def chunks(lst: list, n: int):
        """Yields successive n-sized chunks from a list.

        Parameters
        ----------
        lst : list
            list
        n : int
            chunck size

        Yields
        -------
        list
            chunk
        """
        for i in range(0, len(lst), n):
            yield lst[i : i + n]

    def parse_table_to_madx_line_file(self, filename: str):
        """Method to convert table to madx line def lattice file string and write to file.

        Parameters
        ----------
        filename : str
            file to write to
        """
        save_string(self.parse_table_to_madx_line_string(), filename)

    def get_quad_strengths(self) -> dict:
        """Method to return quadrupole strengths as a dict.

        Returns:
        --------
        dict
            dictionary with quadrupole strengths
        """
        return (
            self.table.loc[self.table.family == "QUADRUPOLE", ["name", "K1"]]
            .set_index("name", drop=True)
            .to_dict()["K1"]
        )

    def get_sext_strengths(self) -> dict:
        """Method to return sextupole strengths as a dict

        Returns:
        --------
        dict
            dictionary with sextupole strengths
        """
        if "SEXTUPOLE" in self.table.family.values:
            return (
                self.table.loc[self.table.family == "SEXTUPOLE", ["name", "K2"]]
                .set_index("name", drop=True)
                .to_dict()["K2"]
            )
        else:
            return {}

    def load_strengths_to_table(self, strdc: dict, col: str) -> None:
        """Method to load a dictionary with strength settings to the table.
        The col attribute is where the strengths will be loaded to.

        Parameters
        ----------
        strdc : dict
            strength settings dict
        col : str
            column where to write the settings
        """
        self.history.put((deepcopy(self.name), deepcopy(self.len), deepcopy(self.table)))

        for k, v in strdc.items():
            self.table.loc[self.table["name"] == k, col] = v

    def compare_seq_center_positions(self, seqfile2: str) -> Tuple[pd.DataFrame]:
        """Compares the center positions of elements of the lattice in the table
        with another lattice table.

        Parameters
        ----------
        seqfile2 : str
            MADX sequence file of the second lattice.add()

        Returns
        -------
        Tuple[pd.DataFrame]
            returns two dataframes, the first is with equal positions, the second with the different positions
        """
        # assert os.path.isfile(seqfile1)
        assert os.path.isfile(seqfile2)

        # name1, len1, df1 = parse_from_madx_sequence_file(seqfile1)
        name2, len2, df2 = parse_from_madx_sequence_file(seqfile2)

        table1 = self.table[["name", "pos"]]
        table2 = df2[["name", "pos"]]

        eq = pd.merge(table1, table2, on=["pos"], how="inner")
        diff = table1[~table1["pos"].isin(table2["pos"])]

        return eq, diff

    def update_table(self):
        """Reload the table from the builder table."""
        # updated history
        self.history.put((deepcopy(self.name), deepcopy(self.len), deepcopy(self.table)))

        self.builder.positions = None
        self.builder.build_table()

        # extract the relavant info
        self.name = self.builder.name
        self.table = self.builder.table

        # length is last element center pos + half the  length
        print("Length has been autoset - check if value is ok - otherwise update it.")
        self.len = (
            self.builder.table.tail(1)["at"].values[-1]
            + self.builder.table.tail(1)["L"].values[-1] / 2.0
        )

    def undo(self):
        """Undo previous change."""
        if not self.history.empty():
            old = self.history.get()
            self.name, self.length, self.table = old
        else:
            print("No previous states available")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['latticeadaptor',
 'latticeadaptor.lark',
 'latticeadaptor.mappings',
 'latticeadaptor.parsers',
 'latticeadaptor.utils']

package_data = \
{'': ['*']}

install_requires = \
['LatticeJSON>=0.1.6,<0.2.0',
 'cpymad>=1.8.1,<2.0.0',
 'ipykernel>=5.5.5,<6.0.0',
 'ipython>=7.23.1,<8.0.0',
 'lark>=0.11.3,<0.12.0',
 'latticeconstructor>=0.2.3,<0.3.0',
 'matplotlib>=3.4.1,<4.0.0',
 'nbsphinx>=0.8.5,<0.9.0',
 'pandas>=1.2.4,<2.0.0',
 'pandoc>=1.0.2,<2.0.0',
 'termcolor>=1.1.0,<2.0.0']

setup_kwargs = {
    'name': 'latticeadaptor',
    'version': '0.3.9',
    'description': '<Enter a one-sentence description of this project here.>',
    'long_description': '==============\nlatticeadaptor\n==============\n\n.. image:: https://readthedocs.org/projects/latticeadaptor/badge/\n    :alt: Documentation Status\n    :scale: 100%\n    :target: http://latticeadaptor.readthedocs.io\n\n\n.. image:: https://mybinder.org/badge_logo.svg\n :target: https://mybinder.org/v2/gh/tomerten/latticeadaptor/main?filepath=docs%2Fnotebook%2FTutorial.ipynb\n\n\nThis package is used to convert a table of accelerator lattice elements and their arguments \nto various accelerator lattice formats. Currently covered are:\n\n- MadX (Sequence and Line format)\n- Elegant\n- Tracy\n\n* Free software: MIT license\n* Documentation: https://latticeadaptor.readthedocs.io.\n\n\nFeatures\n--------\n\n* Fast lattice comparison by plotting element locations against eachother.\n* Fast settings compare by comparing magnet settings.\n* Quick twiss plotting.\n* Quick lattice element plotting.\n* Conversion from table to standard formats.\n\nTutorial on Binder\n------------------\n\nhttps://mybinder.org/v2/gh/tomerten/latticeadaptor/main?filepath=docs%2Fnotebook%2FTutorial.ipynb\n\n',
    'author': 'Tom Mertens',
    'author_email': 'your.email@whatev.er',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/tomerten/latticeadaptor',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-

"""
Module latticeadaptor.parsers 
=================================================================

A module containg accelerator lattice parsers for converting between formats.

"""

# your imports here ...
from abc import ABC
from json import load
from pathlib import Path
from typing import Tuple

import numpy as np
import pandas as pd
from lark import Lark, Transformer, v_args
from lark.exceptions import LarkError

# ==============================================================================
# SET BASE DIR
# LOAD MAPPINGS BETWEEN THE DIFFERENT FORMATS
# ==============================================================================

BASE_DIR = Path(__file__).resolve().parent

with (BASE_DIR / "../lark/Madx.lark").open() as file:
    MADX_PARSER = Lark(file, parser="lalr", maybe_placeholders=True)
    file.seek(0)

# MADX
with (BASE_DIR / "../mappings/madx_columns.json").open() as file:
    MADX_ATTRIBUTES = load(file)

# ELEGANT
with (BASE_DIR / "../mappings/elegant_columns.json").open() as file:
    ELEGANT_ATTRIBUTES = load(file)

with (BASE_DIR / "../mappings/elegant_element_map.json").open() as file:
    TO_ELEGANT_ELEMENTS = load(file)

with (BASE_DIR / "../mappings/elegant_attribute_map.json").open() as file:
    TO_ELEGANT_ATTR = load(file)

# TRACY
with (BASE_DIR / "../mappings/tracy_columns.json").open() as file:
    TRACY_ATTRIBUTES = load(file)

with (BASE_DIR / "../mappings/tracy_element_map.json").open() as file:
    TO_TRACY_ELEMENTS = load(file)

with (BASE_DIR / "../mappings/tracy_attribute_map.json").open() as file:
    TO_TRACY_ATTR = load(file)


def save_string(string: str, file: str) -> None:
    """Quick method to save string to file.add()

    Parameters
    ----------
    string : str
        String to be saved to file.add()
    file : str
        filename where to save the string
    """
    with open(file, "w") as f:
        f.write(string)


@v_args(inline=True)
class AbstractSequenceFileTransformer(ABC, Transformer):
    """Lark Transformer Class"""

    def transform(self, tree):
        """Method to transform the parsed data"""
        self.elements = []
        self.seq = None
        self.name = None
        self.length = 0.0
        super().transform(tree)
        return self.seq, self.elements, self.name, self.length

    int = int
    float = float
    word = str
    neg = lambda self, item: -item
    number = float
    name = lambda self, item: item.value.upper()
    string = lambda self, item: item[1:-1]

    def element(self, name, type_, *attributes):
        # self.elements[name.upper()] = {**{'family': type_.upper()}, **dict(attributes)}
        self.elements.append(
            {**{"name": name.upper(), "family": type_.upper()}, **dict(attributes)}
        )
        return name

    def attribute(self, name, value):
        return name.upper(), value

    def seq_element(self, name, value):
        return name.upper(), value

    def sequence(self, name, *attr):
        self.name = name
        self.length = attr[0][1]
        return name, attr

    def seq_elements(self, *attr):
        self.seq = attr

    def true(self, *attr):
        return True


@v_args(inline=True)
class MADXTransformer(AbstractSequenceFileTransformer):
    pass


from typing import Tuple, Union

from latticeconstructor.parse import parse_elegant, parse_madx
from latticejson.convert import FROM_ELEGANT, TO_MADX

_CONVERSION_DICT = {
    "KQUAD": "QUADRUPOLE",
    "KSEXT": "SEXTUPOLE",
    "KOCT": "OCTUPOLE",
    "DRIF": "DRIFT",
    "RFCA": "RFCAVITY",
    "CSBEND": "SBEND",
    "MONI": "MONITOR",
    "WATCH": "MARKER",
    "EVKICK": "VKICKER",
    "EHKICK": "HKICKER",
    "MARK": "MARKER",
}


def parse_from_string(string: str, ftype: str = "lte") -> Tuple[Union[str, None], dict, list]:
    """Method to parse an elegant lattice string to name, defintions and ordered
    lattice element list.

    Parameters
    ----------
    string : str
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    Lattice string.

    ftype: optional, str, default = 'lte'
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    format type

    Returns
    -------
    Tuple[Union[str,None],dict, list]
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    name, element definitions dict, element ordered list

    """
    assert ftype.lower() in ["lte", "madx"]

    # use latticejson parser
    if ftype == "lte":
        latdata = parse_elegant(string)
    else:
        latdata = parse_madx(string)

    # create command dict for later use
    cdict = {}
    for command in latdata.get("commands"):
        try:
            cdict[command[0]] = command[1] if len(command[1]) == 1 else list(command[1])
        except Exception:
            cdict[command[0]] = ""

    # get the element definitions dict
    definitions = latdata.get("elements", [])

    # convert to madx names
    definitions = {}
    for name, (_type, attributes) in latdata.get("elements", []).items():
        madtype = _type.upper()
        if ftype == "lte":
            # if no matching type make a marker
            try:
                madtype = TO_MADX[FROM_ELEGANT[_type.lower()]].upper()
            except Exception:
                madtype = _CONVERSION_DICT[_type.upper()]

        definitions[name.upper()] = {
            **{"family": madtype},
            **{k.upper(): v for k, v in attributes.items()},
        }

    # get the lattice sub-lattice dicts
    lattice = latdata.get("lattices", [])

    # method to flatten the sublattices
    def flatten(sublat_dict):
        def _walker(k, lattices=sublat_dict):
            for child in lattices[k]:
                if child in lattices:
                    yield from _walker(child)
                else:
                    yield child

        return _walker(list(sublat_dict.keys())[-1])

    # load ordered list to lattice
    # TODO: requested madx sequence lark parser to be merged in main branch so the
    # hack below can be avoided
    if bool(lattice):
        lattice = list(flatten(lattice))
        lattice = [el.upper() for el in lattice]
    else:
        cdict.pop("ENDSEQUENCE")
        lattice = [el.upper() for el in list(cdict.keys())]

    # attempt to load lattice name
    lattice_name = cdict.get("use", None)

    return lattice_name, definitions, lattice


def parse_from_madx_sequence_string(string: str) -> Tuple[str, float, pd.DataFrame]:
    """Method to parse madx seq string to table format

    :param str string: Madx sequence string
    :returns: name, length, dataframe containing the lattice elements
    """
    # use lark to parse the string
    tree = MADX_PARSER.parse(string)
    positions, elements, name, length = MADXTransformer().transform(tree)

    # read the positions of the elements ('at' in the seq file)
    if positions is not None:
        dfpos = pd.DataFrame.from_records(positions, columns=["name", "pos"])
    else:
        dfpos = pd.DataFrame()

    # if not bare sequence file
    if elements:
        dfel = pd.DataFrame(elements)

        # if positions are available merge the tables
        if positions:
            df = dfpos.merge(dfel, on="name").sort_values(by="pos")
            df.loc[df.L.isna(), "L"] = 0
            df["at"] = df["pos"]
            return name, length, df
        else:
            return name, length, dfel

    # if seq file is bare print warning and return only
    # pos table as table
    print("Warning: bare lattice only positions returned")

    return name, length, dfpos


def parse_from_madx_sequence_file(filename: str) -> Tuple[str, float, pd.DataFrame]:
    """Method to parse madx seq from file to table format.

    :param str filename: filename of the file containing the sequence.
    """
    with open(filename, "r") as f:
        string = f.read()

    return parse_from_madx_sequence_string(string)


# ==============================================================================
# Table parsers
# ==============================================================================
def _parse_table_to_madx_definitions(df: pd.DataFrame) -> str:
    """
    Method to parse table to MADX sequence file definitions.

            :param pd.DataFrame df: table to parse
            :returns: Madx string for the element defintions.
    """
    # init output
    text = """"""

    df = df.drop(columns=["pos", "at"], errors="ignore")
    df = df.drop_duplicates()
    # loop over the rows of the frame
    for _, row in df.iterrows():
        # get the element family to check against allowed attrs
        keyword = row["family"]

        # get allowed attrs - to distinguish madx from elegant columns
        allowed_attrs = MADX_ATTRIBUTES[keyword].keys()

        # init line
        line = ""

        # name and element type
        line += "{:16}: {:12}, ".format(row["name"], keyword)

        # remove non attrs from columns
        row = row.drop(
            ["name", "at", "family", "end_pos", "sector", "pos"], errors="ignore"
        ).dropna()

        # add allowed madx attributes
        if len(allowed_attrs) > 0:
            attr_line = (
                ", ".join(
                    [
                        "{}:={}".format(c, row[c])
                        if c in allowed_attrs and c != "NO_CAVITY_TOTALPATH"
                        else "{}={}".format(c, str(row[c]).lower())
                        if c == "NO_CAVITY_TOTALPATH"
                        else ""
                        for c in row.index
                    ]
                )
                + ";\n"
            )
        else:
            attr_line = ";\n"
            line = line[:-2]

        line += attr_line

        # add line to text
        text += line

    return text


def _parse_table_to_madx_sequence_part(name: str, length: float, df: pd.DataFrame) -> str:
    """
    Method to parse a table to the MADX sequence part.

            :param str name: lattice name
            :param float length: lattice length
            :param pd.DataFrame df: table to parse
            :returns: Madx sequence string with positions of the elements.
    """
    # start the sequence definition
    text = "{}: SEQUENCE, L={};\n".format(name, length)

    # loop over the table rows
    for _, row in df.iterrows():
        line = "{:11}, at = {:12.6f};\n".format(row["name"], row["at"])
        text += line

    # close the sequence definition
    text += "ENDSEQUENCE;"

    return text


def parse_table_to_madx_sequence_string(name: str, length: float, df: pd.DataFrame) -> str:
    """
    Method to parse table to MADX sequence.


            :param str name: lattice name
            :param float length: lattice length
            :param pd.DataFrame df: table to parse
            :returns: Madx sequence string

    """
    # parse the element definitions
    text = _parse_table_to_madx_definitions(df)

    # parse the element positions
    text += _parse_table_to_madx_sequence_part(name, length, df)

    return text


def parse_table_to_madx_sequence_file(
    name: str, length: float, df: pd.DataFrame, filename: str
) -> None:
    """Method to parse table to madx sequence and save in file.

    :param str name: lattice name
    :param float length: lattice length
    :param pd.DataFrame df: table to parse
            :param str filename: filename of the file
    """
    save_string(parse_table_to_madx_sequence_string(name, length, df), filename)


def parse_table_to_madx_install_str(name: str, df: pd.DataFrame) -> str:
    """
    Method to parse table to MADX SEQEDIT INSTALL string.
    This can be saved to file to load via CALL or used
    directly as MADX input string using MADX().input(str) from
    cpymad package.

            :param str name: lattice name
            :param pd.DataFrame df: table with elements to install
            :returns: Madx install str
    """

    # start sequence edit
    text = "USE, SEQUENCE={};\n".format(name)
    text += "SEQEDIT, SEQUENCE = {};  \nFLATTEN;\n".format(name)
    for _, row in df.iterrows():
        line = "INSTALL, ELEMENT = {:16}, AT = {:12.6f};\n".format(row["name"], row["at"])
        text += line

    # end sequence edit
    text += "FLATTEN;\nENDEDIT;"

    return text


def parse_table_to_madx_remove_str(name: str, df: pd.DataFrame) -> str:
    """
    Method to parse a seq table to MADX SEQEDIT REMOVE string.
    This can be saved to file to load via CALL or used directly
    as MADX input string using MADX().input(str) from the cpymad
    package.

            :param str name: lattice name
            :param pd.DataFrame df: table with elements to remove
            :returns: Madx remove string
    """
    # start sequence edit
    text = "USE, SEQUENCE={};\n".format(name)
    text += "SEQEDIT, SEQUENCE = {};  \nFLATTEN;\n".format(name)
    for _, row in df.iterrows():
        line = "REMOVE, ELEMENT = {:16};\n".format(row["name"])
        text += line

    # end sequence edit
    text += "FLATTEN;\nENDEDIT;"

    return text


def parse_table_to_elegant_string(name: str, df: pd.DataFrame) -> str:
    """
    Method to transform the MADX seq table to an Elegant lte file.add()

            :param str name: lattice name
            :param pd.DataFrame df: table
            :returns: elegant lattice string
    """
    # init output
    text = """"""
    lattice_template = "{}: LINE=({})".format
    # element_template = "{}: {}, {}".format

    df = df.drop(columns=["pos", "at"], errors="ignore")
    lattice_elements = ", ".join(list(df["name"].values))
    lattice = lattice_template(name, lattice_elements)

    df = df.drop_duplicates()
    # loop over the rows of the frame
    for _, row in df.iterrows():
        # get the element family to check against allowed attrs
        keyword = TO_ELEGANT_ELEMENTS[row["family"]]

        # get allowed attrs - to distinguish madx from elegant columns
        # print(keyword)
        # print(TO_ELEGANT_ATTR)
        allowed_attrs = ELEGANT_ATTRIBUTES[keyword]

        line = ""

        # name and element type
        line += "{:16}: {:12}, ".format(row["name"], keyword)

        # remove non attrs from columns
        row = row.drop(["name", "at", "family", "end_pos", "sector"], errors="ignore").dropna()
        # nrow = [TO_ELEGANT_ATTR[c] for c in row.index if (TO_ELEGANT_ATTR[c] != "")]
        nrow = [TO_ELEGANT_ATTR[c] for c in row.index if c in allowed_attrs]
        # print(row.index)

        # add allowed madx attributes
        if len(allowed_attrs) > 0 and len(nrow) > 0:
            attr_line = (
                ", ".join(
                    [
                        "{}={:16.12f}".format(c, row[c])
                        if TO_ELEGANT_ATTR[c] in allowed_attrs and not isinstance(row[c], str)
                        else "{}={:16}".format(c, row[c])
                        if TO_ELEGANT_ATTR[c] in allowed_attrs
                        else ""
                        for c in nrow
                        # if TO_ELEGANT_ATTR[c] in allowed_attrs
                    ]
                )
                + "\n"
            )
        else:
            attr_line = "\n"
            line = line[:-2]

        line += attr_line

        # add line to text
        text += line
        # print(text)

    text += "\n\n"
    text += lattice

    return text


def parse_table_to_elegant_file(name: str, df: pd.DataFrame, filename: str) -> None:
    """Method to parse table to elegant lattice and save to file.

    :param str name: lattice name
    :param pd.DataFrame df: table
    """
    save_string(parse_table_to_elegant_string(name, df), filename)


def parse_table_to_tracy_string(name: str, df: pd.DataFrame) -> str:
    """
    Method to transform the MADX seq table to tracy lattice string.

            :param str name: lattice name
            :param pd.DataFrame df: table
            :returns: tracy lattice string
    """

    # init output
    text = """"""
    template_marker = "{}: Marker;".format
    template_bpm = "{}: Beam Position Monitor;".format
    template_drift = "{}: Drift, {};".format
    template_bend = "{}: Bending, {}, N = Nbend, Method = 4;".format
    template_quad = "{}: Quadrupole, {}, N = Nquad, Method = 4;".format
    template_sext = "{}: Sextupole, {}, N = Nsext, Method = 4;".format
    template_oct = "{}: Multipole, L = {}, HOM = (4,{}/6.0,0.0), N = Nsext, Method = 4;".format
    template_cav = "{}: Cavity, {};".format

    latname = name
    lattice_template = "{}: "
    n_elem = 10
    lattice_elements = list(df["name"].values)
    n = len(lattice_elements)
    if n >= n_elem:
        lattice_template += "\n "
    for k in range(2, n + 2):
        if (k - 1) % (n_elem + 1) == 0:
            lattice_template += "\n "
        lattice_template += lattice_elements[k - 2]
        if k < n + 1:
            lattice_template += ", "
        else:
            lattice_template += ";"
    # element_template = "{}: {}, {}".format

    df = df.drop(columns=["pos", "at"], errors="ignore")
    lattice = lattice_template.format(latname)

    df = df.drop_duplicates()
    # loop over the rows of the frame
    for _, row in df.iterrows():
        # get the element family to check against allowed attrs
        keyword = TO_TRACY_ELEMENTS[row["family"]]
        name = row["name"]

        # get allowed attrs - to distinguish madx from elegant columns
        allowed_attrs = TRACY_ATTRIBUTES[keyword]
        # print(allowed_attrs)

        # line = ""
        # name and element type
        # line += "{:16}: {:12}, ".format(row["name"], keyword)

        # remove non attrs from columns
        row = row.drop(["name", "at", "family", "end_pos", "sector"], errors="ignore").dropna()
        # print(row.index)
        # update the indices

        try:
            row.index = [TO_TRACY_ATTR[c] for c in row.index if TO_TRACY_ATTR[c] != ""]
            # row.index = [TO_TRACY_ATTR[c] for c in row.index if c in allowed_attrs]
            nrow = row.index
        except:
            nrow = [TO_TRACY_ATTR[c] for c in row.index if TO_TRACY_ATTR[c] != ""]
            # nrow = [TO_TRACY_ATTR[c] for c in row.index if c in allowed_attrs]

        if keyword == "bpm":
            line = template_bpm(name)
        elif keyword == "marker":
            line = template_marker(name)
        elif keyword == "drift":
            line = template_drift(name, "L = {}".format(row["L"]))
        elif keyword == "bend":
            new_row = {"L": row["L"]}
            new_row["T"] = np.degrees(row["T"])

            if "Roll" in nrow:
                new_row["Roll"] = np.degrees(row.get("Roll", None))

            if "Gap" in nrow:
                new_row["Gap"] = 4.0 * row.Gap * row.loc_fint

            if "T1" in nrow:
                new_row["T1"] = np.degrees(row.T1)
            if "T2" in nrow:
                new_row["T2"] = np.degrees(row.T2)
            if "K" in nrow:
                new_row["K"] = row.K

            line = template_bend(
                name,
                ", ".join(
                    [
                        "{} = {:17.15f}".format(k, v)
                        if k not in ["L", "K"]
                        else "{} = {:8.6f}".format(k, v)
                        for k, v in new_row.items()
                    ]
                ),
            )
            # line += ", N = Nbend, Method = 4;"

        elif keyword == "quad":
            line = template_quad(
                name,
                ", ".join(
                    ["{} = {:8.6f}".format(k, v) for k, v in row.items() if k in allowed_attrs]
                ),
            )
            # line += ", N = Nquad, Method = 4;"

        elif keyword == "sext":
            line = template_sext(
                name,
                ", ".join(
                    [
                        "{} = {:8.6f}".format(k, v) if k != "K" else "{} = {}/2.0".format(k, v)
                        for k, v in row.items()
                        if k in allowed_attrs
                    ]
                ),
            )
            # line += ", N = Nsext, Method = 4;"
        elif keyword == " oct1":
            line = template_oct(name, row["L"], row["K"])

        elif keyword == "cavity":
            new_row = {"L": row["L"]}
            new_row["Frequency"] = row.get("Frequency", 0.0)
            new_row["Voltage"] = row.get("Voltage", 0.0)
            new_row["phi"] = row.get("phi", 0)

            line = template_cav(
                name,
                ", ".join(
                    [
                        "{} = {:17.15f}".format(k, float(v))
                        if k not in ["L"]
                        else "{} = {:8.6f}".format(k, float(v))
                        for k, v in new_row.items()
                    ]
                ),
            )

        line += "\n"

        # add line to text
        text += line
        # print(text)
    text += "\n\n"
    text += lattice

    text += "\n\n"
    text += "ring: {};\n\n".format(latname)
    text += "cell: ring, symmetry = 1;"
    text += "\n\nend;"

    return text


def parse_table_to_tracy_file(name: str, df: pd.DataFrame, filename: str) -> None:
    """Method to transform the MADX seq table to tracy lattice and write to file.

    :param str name: lattice name
    :param pd.DataFrame df: table
    :param str filename: file where to save to
    """
    save_string(parse_table_to_tracy_string(name, df), filename)
